unit Main;

{

Memory 3D - Full Source Code Delphi Game
http://zarko-gajic.iz.hr/memory-3d-full-source-code-delphi-game/


Icons courtesy of: https://www.freepik.com/makyzz


}


interface

uses
  MemoryGame, System.Generics.Defaults, System.Generics.Collections, ShellApi,
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, Vcl.StdCtrls,
  Vcl.Buttons, Vcl.Imaging.jpeg, Vcl.Imaging.pngimage;

type
  TMainForm = class(TForm)
    pnlGame: TPanel;
    pnlGameField: TPanel;
    pnlNewGame: TPanel;
    btnNewGame: TButton;
    pnlCurrentGame: TPanel;
    ledPairs: TLabeledEdit;
    ledPlayers: TLabeledEdit;
    gameGrid: TGridPanel;
    lblCurrentPlayerName: TLabel;
    txtCurrentPlayer: TStaticText;
    lblCurrentPlayerPairs: TLabel;
    txtStatistics: TStaticText;
    lbPlayerStatistics: TListBox;
    ledPlanes: TLabeledEdit;
    pnlPlanes: TPanel;
    btnPrevPlane: TButton;
    bntNextPlane: TButton;
    lblCurrentPlane: TLabel;
    imgImagesGrid100: TImage;
    imgQuestionField: TImage;
    pnlGameGrid: TPanel;
    lblWeb: TLabel;
    gbOptions: TGroupBox;
    chkShowPlaneInfo: TCheckBox;
    chkShowClaimedInfo: TCheckBox;
    pnlImgs: TPanel;
    procedure btnNewGameClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure bntNextPlaneClick(Sender: TObject);
    procedure btnPrevPlaneClick(Sender: TObject);
    procedure lblWebMouseEnter(Sender: TObject);
    procedure lblWebMouseLeave(Sender: TObject);
    procedure lblWebClick(Sender: TObject);
    procedure chkShowPlaneInfoClick(Sender: TObject);
    procedure chkShowClaimedInfoClick(Sender: TObject);
  private
    //determined by the images grid - how many single images do we have
    fClipImage : TImage;
    fClipColumns : integer;
    fClipRows : integer;
    fMaxPairs : integer;

    property ClipImage : TImage read fClipImage;
    property ClipColumns : integer read fClipColumns;
    property ClipRows : integer read fClipColumns;
    property MaxPairs : integer read fMaxPairs;
  private
    PictureClip :TPicture;

    fMGame : TMemoryGame;
    function GetMGame: TMemoryGame;
    property MGame : TMemoryGame read GetMGame;

    procedure FieldClaimed(Sender : TObject; const mField : TMField);
    procedure OpenField(Sender : TObject; const mField : TMField);
    procedure CloseField(Sender : TObject; const mField : TMField);
    procedure FieldsPaired(Sender : TObject; const mField1, mField2 : TMField);
    procedure NextPlayer(Sender : TObject; const player : TPlayer);
    procedure GameOver(Sender : TObject; const player : TPlayer);
    procedure PlaneChanged(Sender : TObject);
    procedure GameCreated(Sender : TObject);
    procedure UpdatePlayerStatistics;

    function GetFieldGraphics(const fieldIndex : integer) : TPicture;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

{ TMainForm }

procedure TMainForm.bntNextPlaneClick(Sender: TObject);
begin
  MGame.NextPlane;
end;

procedure TMainForm.btnNewGameClick(Sender: TObject);
var
  newGamePairs, newGamePlayers, newGamePlanes : integer;
  newGameFieldValueShift : integer;
begin
  Randomize;

  newGamePairs := StrToInt(ledPairs.Text);
  newGamePlayers := StrToInt(ledPlayers.Text);;
  newGamePlanes := StrToInt(ledPlanes.Text);;

  fClipImage := imgImagesGrid100;
  fClipColumns := 10;
  fClipRows := 10;
  fMaxPairs := fClipColumns * fClipRows;

  if newGamePairs > MaxPairs then
  begin
    newGamePairs := MaxPairs;
    ledPairs.Text := newGamePairs.ToString;
  end;

  newGameFieldValueShift := Random(MaxPairs - newGamePairs + 1);

  try
    MGame.NewGame(newGamePairs, newGamePlayers, newGamePlanes, newGameFieldValueShift);
  except on E:Exception do
    begin
      MessageDlg(E.Message, mtError, [mbOk], -1);
      Exit;
    end;
  end;
end;

procedure TMainForm.btnPrevPlaneClick(Sender: TObject);
begin
  MGame.PrevPlane;
end;

procedure TMainForm.chkShowClaimedInfoClick(Sender: TObject);
begin
  chkShowPlaneInfoClick(chkShowClaimedInfo);

end;

procedure TMainForm.chkShowPlaneInfoClick(Sender: TObject);
var
  i : integer;
  mField : TMField;
  pnl : TPanel;
  img : TImage;
  lbl : TLabel;
begin
  if MGame.Fields.Count = 0 then Exit; //not started

  for i := 0 to -1 + gameGrid.ControlCount do
  begin
    mField := MGame.Fields[i + MGame.FirstFieldIndexOnCurrentPlane];
    if mField.Host = nil then Exit;

    pnl := TPanel(gameGrid.Controls[i]);
    img := TImage(pnl.Controls[0]);
    lbl := TLabel(pnl.Controls[1]);

    if chkShowPlaneInfo.Checked then
    begin
      lbl.Caption := Format('%d / %d / %d', [i + 1, MGame.CurrentPlane, MGame.PlanesCount]);
      lbl.Visible := MGame.PlanesCount > 1;
    end
    else
      lbl.Visible := false;

    mField.Host := img;

    if (mField.Player <> nil) AND chkShowClaimedInfo.Checked then
    begin
      lbl.Caption := Format('%s', [mField.Player.Name]);
      lbl.Visible := true;
    end
  end;

//  pnlGameGrid.Refresh;
end;

procedure TMainForm.CloseField(Sender : TObject; const mField: TMField);
begin
  TImage(mField.Host).Picture.Assign(imgQuestionField.Picture)
end;

procedure TMainForm.FieldClaimed(Sender : TObject; const mField: TMField);
begin
  TPanel(TImage(mField.Host).Parent).BevelKind := bkNone;
  TPanel(TImage(mField.Host).Parent).Refresh;
  Sleep(200);
  if mField.Player = nil then
    TPanel(TImage(mField.Host).Parent).BevelKind := bkTile
  else
    TPanel(TImage(mField.Host).Parent).BevelKind := bkNone;
  TPanel(TImage(mField.Host).Parent).Refresh;
end;


procedure TMainForm.FieldsPaired(Sender : TObject; const mField1, mField2: TMField);
begin
  if Assigned(mField1.Host) then
  begin
    TPanel(TImage(mField1.Host).Parent).BevelKind := bkNone;
    if chkShowClaimedInfo.Checked then
    begin
      TLabel(TPanel(TImage(mField1.Host).Parent).Controls[1]).Caption := Format('%s', [mField1.Player.Name]);
    end;
  end;
  if Assigned(mField2.Host) then
  begin
    TPanel(TImage(mField2.Host).Parent).BevelKind := bkNone;
    if chkShowClaimedInfo.Checked then
    begin
      TLabel(TPanel(TImage(mField2.Host).Parent).Controls[1]).Caption := Format('%s', [mField2.Player.Name]);
    end;
  end;

  lblCurrentPlayerPairs.Caption := Format('Pairs: %d (%d)',[mGame.CurrentPlayer.ClaimedPairs, mGame.CurrentPlayer.Tries]);

  UpdatePlayerStatistics;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  PictureClip := TPicture.Create;

  gameGrid.Caption := '';

  btnNewGame.Click;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  fMGame.Free;
  PictureClip.Free;
end;

procedure TMainForm.GameCreated(Sender: TObject);
var
  img : TImage;
  pnl : TPanel;
  lbl : TLabel;
  i : integer;
begin
  UpdatePlayerStatistics;

  begin //prepare GridPanel UI
    gameGrid.RowCollection.BeginUpdate;
    gameGrid.ColumnCollection.BeginUpdate;

    for i := 0 to -1 + gameGrid.ControlCount do
      gameGrid.Controls[0].Free;

    gameGrid.RowCollection.Clear;
    gameGrid.ColumnCollection.Clear;

    for i := 1 to MGame.GameGridColumns do
      with gameGrid.RowCollection.Add do
      begin
        SizeStyle := ssPercent;
        Value := 100 / MGame.GameGridColumns;
      end;

    for i := 1 to MGame.GameGridRows do
      with gameGrid.ColumnCollection.Add do
      begin
        SizeStyle := ssPercent;
        Value := 100 / MGame.GameGridRows;
      end;

    //create playable hosts on first plane
    for i := 0 to -1 + MGame.FieldsCountPerPlane do
    begin
      pnl := TPanel.Create(self);
      pnl.AlignWithMargins := true;
      pnl.BevelOuter := bvNone;
      pnl.Align := alClient;
      pnl.Parent := gamegrid;

      img := TImage.Create(pnl);
      img.Align := alClient;
      img.Center := true;
      //img.Transparent := true;
      img.Proportional := true;
      img.Stretch := true;
      img.Parent := pnl;

      lbl := TLabel.Create(pnl);
      lbl.Align := alBottom;
      lbl.Alignment := taCenter;
      lbl.AlignWithMargins := true;
      lbl.Parent := pnl;

      img.OnClick := MGame.FieldHostAction;
    end;

    gameGrid.RowCollection.EndUpdate;
    gameGrid.ColumnCollection.EndUpdate;
  end; //prepare GridPanel UI
end;

procedure TMainForm.GameOver(Sender : TObject; const player: TPlayer);
begin
  MessageDlg(Format('We have a winner: %s', [player.Name]), mtInformation, [mbOK], -1);
end;

function TMainForm.GetFieldGraphics(const fieldIndex: integer): TPicture;
var
  clipWidth, clipHeight : Integer;
  srcRect, destRect : TRect;
begin
  //top, left is 0, 0

  clipWidth := ClipImage.Picture.Width div ClipColumns;
  clipHeight := ClipImage.Picture.Height div ClipRows;

  destRect := Rect(0, 0, clipWidth, clipHeight) ;
  srcRect.Left := (fieldIndex mod ClipColumns) * clipWidth;
  srcRect.Top := (fieldIndex div ClipRows) * clipHeight;
  srcRect.Right := srcRect.Left + clipWidth;
  srcRect.Bottom := srcRect.Top + clipHeight;

  PictureClip.Bitmap.Width := clipWidth;
  PictureClip.Bitmap.Height := clipHeight;
  PictureClip.Bitmap.Canvas.CopyRect(destRect, ClipImage.Canvas, srcRect) ;

  result := PictureClip;
end;

function TMainForm.GetMGame: TMemoryGame;
begin
  if fMGame = nil then
  begin
    fMGame := TMemoryGame.Create;

    fMGame.OnOpenField := OpenField;
    fMGame.OnFieldClaimed := FieldClaimed;
    fMGame.OnFieldOpened := FieldClaimed;
    fMGame.OnCloseField := CloseField;
    fMGame.OnFieldsPaired := FieldsPaired;
    fMGame.OnNextPlayer := NextPlayer;
    fMGame.OnGameStart := NextPlayer;
    fMGame.OnGameOver := GameOver;
    fMGame.OnGameCreated := GameCreated;
    fMGame.OnPlaneChanged := PlaneChanged;

  end;
  result := fMGame;
end;

procedure TMainForm.lblWebClick(Sender: TObject);
begin
  ShellExecute(Handle,'open',PChar('http://zarko-gajic.iz.hr'),'','',SW_SHOW);
end;

procedure TMainForm.lblWebMouseEnter(Sender: TObject);
begin
  lblWeb.Font.Style := [fsUnderline];
end;

procedure TMainForm.lblWebMouseLeave(Sender: TObject);
begin
  lblWeb.Font.Style := [];
end;

procedure TMainForm.NextPlayer(Sender : TObject; const player: TPlayer);
begin
  txtCurrentPlayer.Font.Style := [fsBold, fsUnderline];
  txtCurrentPlayer.Refresh;
  Sleep(100);
  txtCurrentPlayer.Font.Style := [fsBold];
  txtCurrentPlayer.Refresh;


  lblCurrentPlayerName.Caption := player.Name;
  lblCurrentPlayerPairs.Caption := Format('Pairs: %d (%d)',[mGame.CurrentPlayer.ClaimedPairs, mGame.CurrentPlayer.Tries]);

  Caption := Format('Memory 3D. Current player: %s', [player.Name]);
end;

procedure TMainForm.OpenField(Sender : TObject; const mField: TMField);
begin
  TImage(mField.Host).Picture.Assign(GetFieldGraphics(mField.Value));
  TImage(mField.Host).Refresh;
end;

procedure TMainForm.PlaneChanged(Sender : TObject);
var
  i : integer;
  mField : TMField;
  pnl : TPanel;
  img : TImage;
  lbl : TLabel;
begin
  lblCurrentPlane.Caption := Format('Current plane: %d of %d planes.', [MGame.CurrentPlane, MGame.PlanesCount]);

  //set field hosts for current plane
  for i := 0 to -1 + gameGrid.ControlCount do
  begin
    mField := MGame.Fields[i + MGame.FirstFieldIndexOnCurrentPlane];

    pnl := TPanel(gameGrid.Controls[i]);
    img := TImage(pnl.Controls[0]);
    lbl := TLabel(pnl.Controls[1]);

    if chkShowPlaneInfo.Checked then
    begin
      lbl.Caption := Format('%d / %d / %d', [i + 1, MGame.CurrentPlane, MGame.PlanesCount]);
      lbl.Visible := MGame.PlanesCount > 1;
    end
    else
      lbl.Visible := false;

    mField.Host := img;

    if mField = MGame.OpenedField then
    begin
      pnl.BevelKind := bkTile;
      TImage(mField.Host).Picture.Assign(GetFieldGraphics(mField.Value))
    end
    else if mField.Player <> nil then
    begin
      pnl.BevelKind := bkNone;
      TImage(mField.Host).Picture.Assign(GetFieldGraphics(mField.Value));

      if chkShowClaimedInfo.Checked then
      begin
        lbl.Caption := Format('%s', [mField.Player.Name]);
        lbl.Visible := true;
      end;
    end
    else
    begin
      pnl.BevelKind := bkTile;
      TImage(mField.Host).Picture.Assign(imgQuestionField.Picture);
      //TImage(mField.Host).Picture.Assign(GetFieldGraphics(mField.Value))
    end;
  end;

  pnlGameGrid.Refresh;

end;

procedure TMainForm.UpdatePlayerStatistics;
var
  player : TPlayer;
begin
  lbPlayerStatistics.Items.BeginUpdate;
  try
    lbPlayerStatistics.Clear;

    for player in MGame.Players do
      lbPlayerStatistics.Items.Add(Format('%s: %d (%d)', [player.Name, player.ClaimedPairs, player.Tries]));
  finally
    lbPlayerStatistics.Items.EndUpdate;
  end;
end;

end.
